/////////////////////////////////////////////////////////////////////////////////////
// This file is part of the POEngine library, part of the PngOptimizer application
// Copyright (C) 2002/2013 Hadrien Nilsson - psydk.org
// This library is distributed under the terms of the GNU LESSER GENERAL PUBLIC LICENSE
// See License.txt for the full license.
/////////////////////////////////////////////////////////////////////////////////////
#ifndef POENGINE_POENGINE_H
#define POENGINE_POENGINE_H

#include "../chustd/chustd.h"
using namespace chustd;

#include "POEngineSettings.h"
#include "POWorkerThread.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
// PNG optimizing engine class
class POEngine
{
public:
	// Color used when notifying progress
	enum TextType
	{
		TT_FilePath,      // black
		TT_RegularInfo,   // gray
		TT_ActionVerb,    // gray :  Creating, Converting, Optimizing...
		TT_SizeInfo,      // gray :  xxx bytes -> yyy bytes
		TT_SizeInfoNum,   // gray :  xxx in "xxx bytes"
		TT_FileEnlarged,  // amber : (103% of the original size)
		TT_FileSizeSame,  // gray :  (100% of the original size)
		TT_FileReduced,   // green : (80% of the original size)
		TT_ActionOk,      // green : (OK)
		TT_ActionFail,    // red :   (KO)
		TT_ErrorMsg,      // red :   could not load image
		TT_Animated,      // blue :  [Animated GIF : ...]
		TT_WelcomeMsg,    // blue :  Drop PNG, GIF, BMP or TGA files here
		TT_BatchDoneOk,   // green: more than two files optimization done
		TT_BatchDoneFail, // red:   more than two files optimization done with errors
		TT_Last
	};

	struct ProgressingArg
	{
		chustd::String     text;
		TextType   textType;
		ProgressingArg() : textType(TT_Last) {}
	};

	chustd::Event1<const ProgressingArg&> Progressing; // Fired during the optimization process

public:
	POEngineSettings m_settings;

public:
	// Optimizes several image files. The Progressing event is fired on a regular basis 
	// with the text to be printed.
	// [in] filePaths : array of absolute file paths
	// [in] joker : file types handled. ex: "*.png" or "*.gif|*.bmp"
	void OptimizeFiles(const chustd::StringArray& filePaths, const chustd::String& joker = "");

	// Optimizes a buffer in memory and outputs the result as a png file.
	// This function ignores chunk options of m_settings
	// [in] ds       : source buffer description
	// [in] filePath : file path of the result
	// Returns true upon success
	bool OptimizeExternalBuffer(const chustd::PngDumpData& ds, const chustd::String& filePath);

	// Optimizes an image file and outputs the result as a png file. 
	// This function ignores m_settings.backupOldPngFiles.
	// [in] filePath    : path of the file to optimize
	// [in] newFilePath : path of the result
	// Returns true upon success
	bool OptimizeFileNoBackup(const chustd::String& filePath, const chustd::String& newFilePath);

	// Gets error explanation when an optimization function fails
	chustd::String GetLastErrorString() const;
	void ClearLastError();

	struct SingleOptiInfo
	{
		uint32 sizeBefore;
		uint32 sizeAfter;

		SingleOptiInfo() { Clear(); }
		void Clear()
		{
			sizeBefore = 0;
			sizeAfter = 0;
		}
	};

	// Optimizes or convert one single file.
	bool OptimizeSingleFile(const chustd::String& filePath, const chustd::String& displayDir, SingleOptiInfo& singleOptiInfo);

	// Optimizes or convert one single file loaded in memory.
	bool OptimizeSingleFileMem(const uint8* imgBuf, int imgSize, uint8* dst, int dstCapacity, int* pDstSize);

	static chustd::Color32 ColorFromTextType(TextType tt);

	// Initializes some structure. To be called before any optimization function to avoid making the first call
	// slower than other calls (first call will do the warmup if it is not done).
	bool WarmUp();

	POEngine();
	virtual ~POEngine();

private:
	// This class allows the storing of the different tries done by POEngine
	// When dumping, we keep the best result of all
	class ResultManager
	{
	public:
		DynamicMemoryFile& GetCandidate(); // Gets the slot to test a new compression flavour
		DynamicMemoryFile& GetSmallest();  // Gets the best slot of all
		
		void Reset();

	private:
		DynamicMemoryFile m_dmf0;
		DynamicMemoryFile m_dmf1;
	};
	
	ResultManager m_resultmgr;
	
	// Last errors
	StringArray m_astrErrors;
	uint64 m_originalFileWriteTime;

	POWorkerThread m_workerThreads[4];

	// Hold either a target file path or a target buffer
	struct OptiTarget
	{
		String filePath;
		void*  buf;
		int    bufCapacity;
		int    size;

		OptiTarget(const String& filePathArg) 
			: filePath(filePathArg) {}
		OptiTarget(void* bufArg, int bufCapacityArg) 
			: buf(bufArg), bufCapacity(bufCapacityArg), size(0) {}
	};

private:
	struct OptiInfo
	{
		int optiCount;
		int errorCount;
		int64 sizeBefore;
		int64 sizeAfter;

		OptiInfo()
		{
			optiCount = 0;
			errorCount = 0;
			sizeBefore = 0;
			sizeAfter = 0;
		}
	};
	void OptimizeFiles(const String& baseDir, const StringArray& filePaths, const String& displayDir, 
	                   const String& joker, OptiInfo& optiInfo);

	bool Optimize(PngDumpData& dd, OptiTarget& target);
	bool OptimizeSingleFileNoBackup(IFile& fileImage, OptiTarget& target);

	// Those functions fill the DynamicMemoryFiles of m_resultmgr
	bool OptimizePaletteMode(PngDumpData& dd);
	bool Optimize24BitsMode(PngDumpData& dd);
	bool Optimize32BitsMode(PngDumpData& dd);
	bool OptimizeGrayScale(PngDumpData& dd);
	bool OptimizeGrayScaleAlpha(PngDumpData& dd);

	int  CanSimplifyGreyAlpha(const PngDumpData& dd) const;
	bool IsBlackAndWhite(const Palette& pal, bool& bShouldSwap);
	bool IsGreyPalette(const Palette& pal);
	bool TryToConvertIndexedToBlackAndWhite(PngDumpData& dd);
	bool TryToConvertIndexedToGreyscale(PngDumpData& dd);
	bool FindUnusedColor(const Buffer& aRgb, uint8& nRed, uint8& nGreen, uint8& nBlue);
	bool FindUnusedColorHardcoreMethod(const uint8* pRgba, int32 nPixelCount, uint8& nRed, uint8& nGreen, uint8& nBlue);
	bool DumpBestResultToFile(OptiTarget& target);

	bool InsertCleanOriginalPngAsResult(IFile& file);

	void AddError(const String& str);

	void PrintSizeChange(int64 sizeBefore, int64 sizeAfter);
	void PrintText(const String& text, TextType textType);
	
	static void BgrToRgb(PngDumpData& dd);
	static void BgraToRgba(PngDumpData& dd);
	static bool Rgb16ToRgb24(PngDumpData& dd);

public:
	// public for unit testing
	bool OptimizeAnimated(const ImageFormat& img, PngDumpData& dd, const String& filePath);
private:
	bool OptimizeAnimated(const ImageFormat& img, PngDumpData& dd, OptiTarget& target);
	bool PerformDumpTries(PngDumpData& ds);

	static void UnpackPixelFrames(PngDumpData& dd);
	static void PackPixelFrames(PngDumpData& dd);
	static bool IsFileExtensionSupported(const String& ext, const String& joker="");
		
	friend class POEngine_IsFileExtensionSupported_Test;
};

#endif
