///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2013 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Console.h"
#include "String.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
namespace chustd {\

class ConsoleColorSetter
{
public:
	void SetColor(WORD col)
	{
		HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
		if( !m_initialized )
		{
			CONSOLE_SCREEN_BUFFER_INFO ci;
			GetConsoleScreenBufferInfo(hStdOut, &ci);
			m_original = ci.wAttributes;
			m_initialized = true;
		}
		int mask = ~(FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE|FOREGROUND_INTENSITY);
		WORD att = (m_original & mask) | col;
		SetConsoleTextAttribute(hStdOut, att);
	}
	void SetNormalColor()
	{
		if( m_initialized )
		{
			HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
			SetConsoleTextAttribute(hStdOut, m_original);
		}
	}
	WORD GetColor()
	{
		HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);
		CONSOLE_SCREEN_BUFFER_INFO ci;
		GetConsoleScreenBufferInfo(hStdOut, &ci);
		int mask = (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE|FOREGROUND_INTENSITY);
		return (ci.wAttributes & mask);
	}
	ConsoleColorSetter()
	{
		m_initialized = false;
		m_original = 0;
	}
	~ConsoleColorSetter()
	{
		SetNormalColor();
	}
private:
	bool m_initialized;
	WORD m_original;
};

static ConsoleColorSetter g_consoleColorSetter;

///////////////////////////////////////////////////////////////////////////////////////////////////
Console::Console()
{

}

Console::~Console()
{

}

void Console::WriteLine(const chustd::String& str)
{
	Write(str + L"\r\n");
}

void Console::Write(const chustd::String& str)
{
#ifdef WIN32
	// Windows functions
	HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);

	DWORD charWrittenCount = 0;
	DWORD length = str.GetLength();
	BOOL bOk = WriteConsoleW(hStdOut, str.GetBuffer(), length, &charWrittenCount, NULL);
	bOk;
#elif CHUOS

#endif
}

bool Console::WaitForInput(chustd::String& str)
{
	static wchar szPrompt[] = L">";
	return WaitForInput(szPrompt, str);
}

// Wait for a string input by the user
// strPrompt : symbol to print before waiting for the string
// strInput : result
bool Console::WaitForInput(const chustd::String& strPrompt, chustd::String& strInput)
{
	Write(strPrompt);

	wchar szBuffer[128];

#ifdef WIN32
	HANDLE hStdIn = GetStdHandle(STD_INPUT_HANDLE);

	DWORD numberOfCharsToRead = (sizeof(szBuffer) / 2) - 1;
	DWORD numberOfCharsRead = 0;

	BOOL bWinOk = ReadConsoleW(hStdIn, szBuffer, numberOfCharsToRead, &numberOfCharsRead, NULL);

	if( !bWinOk )
		return false;
#else

	int32 numberOfCharsRead = 0;
#endif

	// Removes any carriage return or line feed
	int32 i = numberOfCharsRead - 1;
	for(; i >= 0; --i)
	{
		wchar c = szBuffer[i];
		if( c != 13 && c != 10 )
			break;
	}

	int32 realCharCount = i + 1;
	strInput.SetBuffer(szBuffer, realCharCount);
	return true;
}

uint32 Console::WaitForKey()
{
	uint32 char32 = 0;

#ifdef WIN32
	HANDLE hStdIn = GetStdHandle(STD_INPUT_HANDLE);

	// Switch to raw mode (no line input, no echo input)
	DWORD oldMode = 0;
	GetConsoleMode(hStdIn, &oldMode);
	SetConsoleMode(hStdIn, 0);
	
	for(;;)
	{
		DWORD read = 0;
		INPUT_RECORD inputrec;
		if( !ReadConsoleInputW(hStdIn, &inputrec, 1, &read) || (read == 0) )
		{
			char32 = 0;
			break;
		}
		
		if( (inputrec.EventType == KEY_EVENT) && inputrec.Event.KeyEvent.bKeyDown )
		{
			char32 = inputrec.Event.KeyEvent.uChar.UnicodeChar;
			if( char32 != 0 )
			{
				break;
			}
		}
	}
	
	SetConsoleMode(hStdIn, oldMode);
#endif
	return char32;
}

void Console::SetTextColor(Color32 col)
{
	int th1 = 64; // First threshold
	int th2 = 192; // Second threshold

	WORD ccol = 0;
	if( col.r > th1 )
	{
		ccol |= FOREGROUND_RED;
	}
	if( col.g > th1 )
	{
		ccol |= FOREGROUND_GREEN;
	}
	if( col.b > th1 )
	{
		ccol |= FOREGROUND_BLUE;
	}
	if( col.r > th2 || col.g > th2 || col.b > th2 )
	{
		ccol |= FOREGROUND_INTENSITY;
	}
	g_consoleColorSetter.SetColor(ccol);
}

void Console::ResetTextColor()
{
	g_consoleColorSetter.SetNormalColor();
}

Color32 Console::GetTextColor()
{
	Color32 ret;
	WORD ccol = g_consoleColorSetter.GetColor();

	uint8 th = 127;
	if( ccol & FOREGROUND_INTENSITY )
	{
		th = 255;
	}
	if( ccol & FOREGROUND_RED )
	{
		ret.r = th;
	}
	if( ccol & FOREGROUND_GREEN )
	{
		ret.g = th;
	}
	if( ccol & FOREGROUND_BLUE )
	{
		ret.b = th;
	}
	return ret;
}

// Checks if the console was created by this process, and the process was run from an existing console
bool Console::IsOwned()
{
	HWND hConsoleWnd = GetConsoleWindow();
	DWORD dwProcessId = 0;
	GetWindowThreadProcessId(hConsoleWnd, &dwProcessId);
	return GetCurrentProcessId() == dwProcessId;
}

} // namespace chustd
