///////////////////////////////////////////////////////////////////////////////
// This file is part of the chustd library
// Copyright (C) 2005/2013 ChuTeam
// For conditions of distribution and use, see copyright notice in chustd.h
///////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "DateTime.h"
#include "String.h"
#include "StringBuilder.h"
#include "FormatType.h"
#include "FixArray.h"
#include "ScanType.h"
#include "TimeStamp.h"

using namespace chustd;
//////////////////////////////////////////////////////////////////////
Duration::Duration()
{
	m_nStamp = 0;
}

Duration::Duration(int64 chuTimeStamp) : m_nStamp(chuTimeStamp)
{

}

Duration::Duration(int32 days, int8 hours, int8 minutes, int8 seconds, int16 milliseconds)
{
	if( !(-23 <= hours && hours <= 23) )
	{
		hours = 0;
	}

	if( !(-59 <= minutes && minutes <= 59) )
	{
		minutes = 0;
	}

	if( !(-59 <= seconds && seconds <= 59) )
	{
		seconds = 0;
	}

	int64 stamp = days;

	stamp *= 24;
	stamp += hours;

	stamp *= 60;
	stamp += minutes;

	stamp *= 60;
	stamp += seconds;

	stamp *= 1000;
	stamp += milliseconds;

	m_nStamp = stamp;
}

// Returns the time stamp in seconds
int64 Duration::GetStampInSec() const
{
	return m_nStamp / 1000;
}

Duration Duration::FromSeconds(int64 seconds)
{
	Duration dur;
	dur.m_nStamp = 	1000;
	dur.m_nStamp *= seconds;
	return dur;
}

void Duration::Split(int32& days, int8& hours, int8& minutes, int8& seconds, int16& milliseconds) const
{
	int64 next = m_nStamp;

	milliseconds = int16(next % 1000);
	next /= 1000;

	seconds = int8(next % 60);
	next /= 60;

	minutes = int8(next % 60);
	next /= 60;

	hours = int8(next % 24);
	next /= 24;

	days = int32(next);
}

void Duration::FormatFrench(String& str) const
{
	int32 days;
	int8 hours;
	int8 minutes;
	int8 seconds;
	int16 milliseconds;
	Split(days, hours, minutes, seconds, milliseconds);

	String strDays, strHours, strMinutes, strSeconds;
	wchar szBuffer[12];

	if( days > 0 )
	{
		if( days == 1 )
		{
			strDays = L"1 jour ";
		}
		else
		{
			FormatInt32(days, szBuffer);
			strDays = String(szBuffer) + L" jours";
		}
	}

	if( hours > 0 )
	{
		if( hours == 1 )
		{
			strHours = L"1 heure ";
		}
		else
		{
			FormatInt32(hours, szBuffer);
			strHours = String(szBuffer) + L" heures";
		}
	}

	if( minutes > 0 )
	{
		if( minutes == 1 )
		{
			strMinutes = L"1 minute ";
		}
		else
		{
			FormatInt32(minutes, szBuffer);
			strMinutes = String(szBuffer) + L" minutes";
		}
	}
	
	if( seconds > 0 )
	{
		if( seconds == 1 )
		{
			strSeconds = L"1 seconde";
		}
		else
		{
			FormatInt32(seconds, szBuffer);
			strSeconds = String(szBuffer) + L" secondes";
		}
	}

	str = strDays + strHours + strMinutes + strSeconds;
}

void Duration::FormatSmall(class String& str) const
{
	int32 days;
	int8 hours;
	int8 minutes;
	int8 seconds;
	int16 milliseconds;
	Split(days, hours, minutes, seconds, milliseconds);

	String strDays, strHours, strMinutes, strSeconds;
	wchar szBuffer[12];

	if( days > 0 )
	{
		FormatInt32(days, szBuffer);
		strDays = String(szBuffer) + L" j";
	}

	if( hours > 0 )
	{
		FormatInt32(hours, szBuffer);
		strHours = String(szBuffer) + L" h";
	}

	if( minutes > 0 )
	{
		FormatInt32(minutes, szBuffer);
		strMinutes = String(szBuffer) + L" m";
	}
	
	if( seconds > 0 )
	{
		FormatInt32(seconds, szBuffer);
		strSeconds = String(szBuffer) + L" s";
	}

	str = strDays + strHours + strMinutes + strSeconds;
}
//////////////////////////////////////////////////////////////////////
DateTime::DateTime()
{
	m_nYear = 0;
	m_nMonth = 0;
	m_nDay = 0;
	m_nHour = 0;
	m_nMinute = 0;
	m_nSecond = 0;

	m_nStamp = 0;
}

// static
DateTime DateTime::GetNow()
{
	DateTime date;

#ifdef WIN32
	SYSTEMTIME st;
	::GetLocalTime(&st); // OS Dependent
	
	// The FILETIME structure is a 64-bit value representing the number 
	// of 100-nanosecond intervals since January 1, 1601.
	FILETIME ft;
	::SystemTimeToFileTime(&st, &ft);

	int64 stamp = ft.dwHighDateTime;
	stamp <<= 32;
	stamp |= ft.dwLowDateTime;

	date.m_nStamp = stamp / 10000;
	
#elif CHUOS
	date.m_nStamp = chuos_GetLocalTimeStamp();
#endif
	
	date.UpdateFieldsFromStamp();
	return date;
}

void DateTime::UpdateFieldsFromStamp()
{
	uint16 millisecond = 0;
	chustd::SplitChuTimeStamp(m_nStamp, m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond, millisecond);
}

// This function computes a timestamp from the splitted fields
void DateTime::UpdateStampFromFields()
{
	uint16 millisecond = 0;
	m_nStamp = chustd::MakeChuTimeStamp(m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond, millisecond);
}

void DateTime::Set(int16 year, uint8 month, uint8 day, uint8 hour, uint8 minute, uint8 second)
{
	// TODO: think if we should check correctness here
	m_nYear = year;
	m_nMonth = month;
	m_nDay = day;
	m_nHour = hour;
	m_nMinute = minute;
	m_nSecond = second;

	UpdateStampFromFields();
}

// Formats date only as YYYY-MM-DD (ISO 8601)
String DateTime::FormatDate(DateFormat eFormat) const
{
	StringBuilder sbFinal;
	wchar szBuffer[12];

	FormatInt32(m_nYear, szBuffer, 4, L'0');
	sbFinal += szBuffer;

	if( eFormat != dfCompact )
	{
		sbFinal += L"-";
	}
	FormatInt32(m_nMonth, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	if( eFormat != dfCompact )
	{
		sbFinal += L"-";
	}
	FormatInt32(m_nDay, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	return sbFinal.ToString();
}

String DateTime::FormatTime(DateFormat eFormat) const
{
	StringBuilder sbFinal;
	wchar szBuffer[12];

	FormatInt32(m_nHour, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	if( eFormat != dfCompact )
	{
		sbFinal += L":";
	}
	FormatInt32(m_nMinute, szBuffer, 2, L'0');
	sbFinal += szBuffer;


	if( eFormat != dfCompact )
	{
		sbFinal += L":";
	}
	FormatInt32(m_nSecond, szBuffer, 2, L'0');
	sbFinal += szBuffer;

	return sbFinal.ToString();
}

String DateTime::Format(DateFormat eFormat) const
{
	String strDate = FormatDate(eFormat);
	String strTime = FormatTime(eFormat);

	if( eFormat == dfExpandedT || eFormat == dfCompact )
	{
		return strDate + L"T" + strTime;
	}

	return strDate + L" " + strTime;
}

// Gets date as YYYY-MM-DD hh:mm:ss
// Works if last parts are missing, for instance : YYYY-MM
// Returns the number of read members
int32 DateTime::Scan(const String& strDate)
{
	//int count = strDate.Scan("%?-%?-%? %?:%?:%?", m_nYear, m_nMonth, m_nDay, m_nHour, m_nMinute, m_nSecond);
	
	const wchar* psz = strDate.GetBuffer();
	int32 advance = 0;
	
	FixArray<bool, 12> abParseStatus;

	abParseStatus[0] = ScanType(psz, advance, L' ', &m_nYear);
	abParseStatus[1] = psz[advance] == L'-';
	advance++;
	
	abParseStatus[2] = ScanType(psz, advance, L' ', &m_nMonth);
	abParseStatus[3] = psz[advance] == L'-';
	advance++;

	abParseStatus[4] = ScanType(psz, advance, L' ', &m_nDay);
	abParseStatus[5] = psz[advance] == L' ';
	advance++;

	abParseStatus[6] = ScanType(psz, advance, L' ', &m_nHour);
	abParseStatus[7] = psz[advance] == L':';
	advance++;

	abParseStatus[8] = ScanType(psz, advance, L' ', &m_nMinute);
	abParseStatus[9] = psz[advance] == L':';
	advance++;

	abParseStatus[10] = ScanType(psz, advance, L' ', &m_nSecond);
	abParseStatus[11] = true;
	
	int count = 0;
	for(int i = 0; i < abParseStatus.GetSize(); i += 2)
	{
		if( !abParseStatus[i] )
			break;

		count++;
	}

	/////////////////////////////////////////////////////////////////
	// Check fields integrity
	if( count >= 6 && !( m_nSecond <= 59) )
	{
		count = 5;
	}
	if( count >= 5 && !( m_nMinute <= 59) )
	{
		count = 4;
	}
	if( count >= 4 && !( m_nHour <= 23) )
	{
		// Just take the year, the month and the day as the hour is invalid
		count = 3;
	}
	if( count >= 3 && !( 1 <= m_nDay && m_nDay <= 31) )
	{
		// Just take the year and the month as the day is invalid
		count = 2;
	}
	if( count >= 2 && !( 1 <= m_nMonth && m_nMonth <= 12) )
	{
		// Just take the year as the month is invalid
		count = 1;
	}
	/////////////////////////////////////////////////////////////////

	uint8* const apn[] = {null, &m_nMonth, &m_nDay, &m_nHour, &m_nMinute, &m_nSecond };
	int32 iClear = count;

	for(; iClear < 3; ++iClear)
	{
		apn[iClear][0] = 1;
	}

	for(; iClear < 6; ++iClear)
	{
		apn[iClear][0] = 0;
	}

	UpdateStampFromFields();

	return count;
}

void DateTime::AddDays(int count)
{
	int64 oneDay = 24 * 60 * 60;
	oneDay *= 1000;
	m_nStamp += count * oneDay;

	UpdateFieldsFromStamp();
}

void DateTime::SetTime(uint8 hour, uint8 minute, uint8 second)
{
	if( hour > 23 )
	{
		return;
	}

	if( minute > 59 )
	{
		return;
	}

	if( second > 59 )
	{
		return;
	}

	m_nHour = hour;
	m_nMinute = minute;
	m_nSecond = second;

	UpdateStampFromFields();
}

const Duration DateTime::operator - (const DateTime& t2) const
{
	Duration dur;
	dur.m_nStamp = m_nStamp - t2.m_nStamp;

	return dur;
}

bool DateTime::operator > (const DateTime& t2) const
{
	return m_nStamp > t2.m_nStamp;
}

bool DateTime::operator < (const DateTime& t2) const
{
	return m_nStamp < t2.m_nStamp;
}

bool DateTime::operator >= (const DateTime& t2) const
{
	return m_nStamp >= t2.m_nStamp;
}

bool DateTime::operator <= (const DateTime& t2) const
{
	return m_nStamp <= t2.m_nStamp;
}

// Returns the time stamp in seconds
int64 DateTime::GetStampInSec() const
{
	return m_nStamp / 1000;
}

const DateTime& DateTime::operator += (Duration dur)
{
	m_nStamp += dur.m_nStamp;
	UpdateFieldsFromStamp();

	return *this;
}

DateTime DateTime::FromUnixTimeStamp(int64 time)
{
	// 0 == 1970-01-01
	time *= 1000;

	DateTime date;
	date.Set(1970);
	date.m_nStamp += time;

	date.UpdateFieldsFromStamp();
	return date;
}

int64 DateTime::ToUnixTimeStamp() const
{
	DateTime dateBase;
	dateBase.Set(1970);

	int64 dif = m_nStamp - dateBase.m_nStamp;
	dif /= 1000;

	if( dif < 0 )
		dif = 0;

	return dif;
}

int64 DateTime::GetChuTimeStamp() const
{
	return m_nStamp;
}

DateTime DateTime::FromChuTimeStamp(int64 stamp)
{
	DateTime date;
	date.m_nStamp = stamp;
	date.UpdateFieldsFromStamp();
	return date;
}

